/*
* This file is part of simple conversion utility.
*
* Copyright (C) 2018 ZX Cygnus <cygnus1 áť post.cz>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or (at
* your option) any later version.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
* USA
*/

#include <stdio.h>
#include <string.h>
// #include <stdlib.h>

#define VERSION			"0.1"

#define BASLINESTEP		5		// BASIC line number will be incremented by this value
#define BASLINECOMMANDS		10		// how many commands per one BASIC line
#define BASFIRSTLINE		10		// start lines from this number
#define	BASMAXLINE		1000		// stop generating BASIC program if this line will be reached

#define	PARAMETER_LENGTH	5		// 99999/40=2499.975mm it is much more than enough (A0 is 841x1189mm)
						// in reality i expect biggest values around 10400 (260mm), max 11880 (297mm)
						// even for format A3 parameter cannot be bigger than 16800 (420mm), even if
						// there will be zero margins at edge of paper

typedef enum {GL_COMMAND, GL_PARAMETER} hpglstates;
typedef enum {MZXR, ZXBASIC} listmodes;

void add_command_at_basic_line (char *command, int basiccommands)
{
	if (basiccommands == 0) printf (","); else printf (";");
	printf ("%s", command);
}

int converter (char *filename, listmodes listmode, int split, int debug)
{
	hpglstates converter_state;
	FILE *file;
	char c;
	char command[] = "in";			// all commands will be 2 characters long
	int index;				// number of character in command or parameter
	long int hpgl_x,hpgl_y;			// HPGL coordinates
	double mzxr_x, mzxr_y;			// MZXR coordinates
	long int parameter_value;		// computed value from numbers in parameter
	int parameter_index;			// number of parameter (first or second)
	int pen;				// pen state (0 = up)
	int repeat;				// flag - for state machine
	int i;					// universal index in small cycles
	int line;				// line counter, not number of line
	int basicline;				// number of line in generated BASIC program
	int basiccommands;			// basic commands counter

	converter_state = GL_COMMAND;		// expecting first character of HPGL command

	file = fopen(filename,"r");		// read only is enough

	if (file != NULL)
	{
		index = 0;
		pen = 0;			// 0 = pen up, 1 = pen down
		parameter_index = 0;
		repeat = 0;			// if 1, then don't read another character, use current again
		line = 1;
		basicline = BASFIRSTLINE;
		basiccommands = 0;

		if (listmode == ZXBASIC)
		{
			if (debug) printf ("BASIC: ");
			printf ("%i REM generated MZXR drawing", basicline);
			basicline += BASLINESTEP;
			basiccommands = BASLINECOMMANDS;
		}
		else
		{
			if (debug) printf ("OUT: ");
			printf ("pu\n");
		}

		c = fgetc(file);		// read one character before main loop

		while (c != EOF)
		{
			if (basiccommands >= BASLINECOMMANDS)
			{
				// start new (first) BASIC line
				line++;
				if ((line >= 150) && (split == 1))
				{
					// starting next batch of lines, start numbering from start and finish line before new one
					if (pen == 1)
					{
						// at end of splitted part lift pen
						add_command_at_basic_line ("pu", basiccommands);
						if (debug) printf ("\n");
					}
					printf ("\n9999 BEEP 1,1\n");
					printf ("\n--- split here ---\n\n");
					line = 0;
					basicline = BASFIRSTLINE;
				}

				if (debug == 0) printf ("\n");

				if (debug) printf ("BASIC: ");
				printf ("%i PRINT #7", basicline);
				
				basiccommands = 0;

				if ((line == 0) && (split == 1) && (pen == 1))
				{
					// at start of splitted part put pen down if should be
					add_command_at_basic_line ("pd", basiccommands);
					if (debug) printf ("\n");
					basiccommands++;
				}

				if ((basicline == (BASFIRSTLINE + BASLINESTEP)) && (pen == 0))
				{
					// totally first line
					add_command_at_basic_line ("pu", basiccommands);
					if (debug) printf ("\n");
					basiccommands++;
				}

				if (debug) printf ("\n");
				basicline += BASLINESTEP;
			}
			
			switch (converter_state)
			{
				case GL_COMMAND:
					// if byte is not one of letters a-z or A-Z ignore it
					if ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z'))
					{
						if (index == 0)
						{
							if (c >= 'A' && c <= 'Z') c = c + 32;	// lower case
							command [0] = c;
							index = 1;
						}
						else
						{
							if (c >= 'A' && c <= 'Z') c = c + 32;	// lower case
							command [1] = c;
							index = 0;

							if (strcmp(command, "in") == 0)
							{
								if (debug) printf ("DEBUG: Ignoring known command INITIALIZE.\n", command);
								converter_state = GL_COMMAND;
							}
							if (strcmp(command, "sp") == 0)
							{
								if (debug) printf ("DEBUG: Ignoring known command SET PEN.\n", command);
								converter_state = GL_COMMAND;
							}
							if (strcmp(command, "ro") == 0)
							{
								if (debug) printf ("DEBUG: Ignoring known command ROTATE.\n", command);
								converter_state = GL_COMMAND;
							}

							if ((strcmp(command, "pu") == 0) || (strcmp(command, "pd") == 0))
							{
								if (debug) printf ("DEBUG: Command %s - reading parameters\n", command);

								// if pu is pen already up?
								if ((strcmp(command, "pu") == 0) && (pen == 1))
								{
									// pen up
									if (listmode == ZXBASIC)
									{
										if (debug) printf ("BASIC: ");
										add_command_at_basic_line ("pu", basiccommands);
// 										if (basiccommands == 0) printf (","); else printf (";");
// 										printf ("pu");
										if (debug) printf ("\n");
										basiccommands++;
									}
									else
									{
										if (debug) printf ("OUT: ");
										printf ("pu\n");
									}
									pen = 0;
								}

								// if pd is pen already down?
								if ((strcmp(command, "pd") == 0) && (pen == 0))
								{
									// pen down
									if (listmode == ZXBASIC)
									{
										if (debug) printf ("BASIC: ");
										add_command_at_basic_line ("pd", basiccommands);
// 										if (basiccommands == 0) printf (","); else printf (";");
// 										printf ("pd");
										if (debug) printf ("\n");
										basiccommands++;
									}
									else
									{
										if (debug) printf ("OUT: ");
										printf ("pd\n");
									}
									pen = 1;
								}

								// numeric parameters can follow
								converter_state = GL_PARAMETER;
								parameter_value = 0;
								parameter_index = 0;		// commands can have more parameters, typicaly two coordinates
								index = 0;
							}
						}
					}
					else
					{
						// other characters, separators, white spaces, numeric parameters in unknown commands...
						index = 0;
						if (debug) printf("DEBUG: Character %c is not letter, ignoring it.\n", (char)c);
					}
					break;
				case GL_PARAMETER:
					if (c >= '0' && c <= '9')
					{
						if (index < PARAMETER_LENGTH)
						{
							// just add next number into parameter
							parameter_value = (parameter_value * 10) + (c - 48);
							if (debug) printf ("DEBUG: value = %ld\n", parameter_value);
							index++;
						}
						else
						{
							// too long parameter, cannot add another number, ignore whole command
							if (debug) printf ("DEBUG: Numeric parameter is too long! Dropping whole command.\n");
							if (debug) printf ("DEBUG: value = %ld\n", parameter_value);
							converter_state = GL_COMMAND;
							index=0;
						}
					}
					else
					{
						// current character is not number, then parameter is finished or empty
						if (index > 0)
						{
							// parameter finished
							if (debug) printf ("DEBUG: not number, value = %ld\n", parameter_value);
							
							if (parameter_index == 0)
							{
								parameter_index++;
								// X
								hpgl_x = parameter_value;
							}
							else
							{
								parameter_index = 0;
								// Y
								hpgl_y = parameter_value;

								if (listmode == ZXBASIC)
								{
									if (debug) printf ("BASIC: ");
									if (basiccommands == 0) printf (","); else printf (";");
									printf ("ma,VAL\"%g\",VAL\"%g\"", (double)(hpgl_x/40.0), (double)(hpgl_y/40.0));
									if (debug) printf ("\n");
									basiccommands++;
								}
								else
								{
									if (debug) printf ("OUT: ");
									printf ("ma%g,%g\n", (double)(hpgl_x/40.0), (double)(hpgl_y/40.0));
								}
							}
							
							parameter_value = 0;
							index = 0;
							if (c >= 'a' && c <= 'z')
							{
								repeat = 1;
								converter_state = GL_COMMAND;
							}
							if (c == ';') converter_state = GL_COMMAND;
						}
						else
						{
							// parameter is empty, can it be interpretted as zero?  Let's try
							if (debug) printf ("DEBUG: empty parameter\n", parameter_value);
							parameter_value = 0;
							index = 0;
							if (c >= 'a' && c <= 'z')
							{
								repeat = 1;
								converter_state = GL_COMMAND;
							}
							if (c == ';') converter_state = GL_COMMAND;
						}
					}
					break;
				default:
					if (debug) printf ("DEBUG: Unknown mode?\n");
					break;
			}

			// if flag repeat is active, then repeat whole loop with same character, don't read new one
			// because in HPGL can be letter after numeric parametter which is part of following command
			if (repeat == 1)
				repeat = 0;		// reset flag
			else
			{
				c = fgetc(file);
				if ((c == 10) || (c == 13)) c = ' ';
			}
		}
		
		if (listmode == ZXBASIC)
		{
			// add dummy command if last line was started, but remained empty
			if (basiccommands == 0) add_command_at_basic_line ("pu", basiccommands);
			printf ("\n");
		}

		fclose (file);
		return (0);
	}
	else
	{
		printf ("Error: Can't open input file %s.\n",filename);
		fclose (file);
		return (-1);
	}
}

void version ()
{
	printf ("%s \n", VERSION);
}

void help ()
{
	printf ("Usage: hpgl2mzxr [OPTIONS]... [FILENAME]...\n\n");
	printf ("It converts simple hpgl files generated from Inkscape 0.9x\n");
	printf ("into MZXR program for vintage plotter Aritma 0507.\n");
	printf ("\n");
	printf ("If the filename is missing, this help will be printed.\n");
	printf ("  -h, --help    display this help and exit\n");
	printf ("  -b, --basic   print whole ZX BASIC program with MZXR commands, this\n");
	printf ("                program can be compilled with bas2tap utility which is\n");
	printf ("		 part of Fuse Emulator\n");
	printf ("  -s, --split   split BASIC source code in more parts, on every 150th line\n");
	printf ("  -d, --debug   print debug messages\n");

	// -1 = everything on it's own line (1 column)
}

int main(int argc, char *argv[])
{
	int counter;
	int debug = 0;
	int split = 0;
	listmodes listmode = MZXR;
	char *filename;
	int switches;
	int correct;

	switches = 1;
	for (counter = 1; counter<argc; counter++)
	{
//		printf ("++ %i ++ %s \n",counter, argv[counter]);
		correct = 0;

		if (!strcmp(argv[counter],"-h") || (!strcmp(argv[counter],"--help")))
		{
			help ();
			switches++;
			correct = 1;
			return (0);
		}
		if (!strcmp(argv[counter],"-v") || (!strcmp(argv[counter],"--version")))
		{
			version ();
			switches++;
			correct = 1;
			return (0);
		}
 		if (!strcmp(argv[counter],"-d") || (!strcmp(argv[counter],"--debug")))
 		{
 			debug = 1;
 			switches++;
 			correct = 1;
 		}
 		if (!strcmp(argv[counter],"-b") || (!strcmp(argv[counter],"--basic")))
 		{
 			listmode = ZXBASIC;
 			switches++;
 			correct = 1;
 		}
 		if (!strcmp(argv[counter],"-s") || (!strcmp(argv[counter],"--split")))
 		{
 			split = 1;
 			switches++;
 			correct = 1;
 		}
		if ((correct==0) && (counter!=(argc-1)))
		{
			printf ("ERROR: incorrect parametr - \"%s\"\n", argv[counter]);
			printf ("Hint: Filename must be last parameter.\n");
			help ();
			return (-1);
		}
	}

	filename=argv[argc-1];

	if (debug) printf ("DEBUG: switches - %i \n", switches);

	if (switches!=argc-1)
	{
		printf ("ERROR: Filename missing.\n");
		return (-1);
	}
	else
	{
		converter (filename, listmode, split, debug);
		return (0);
	}
}
